#include "def.h"
#include "prot.h"

#define ADR_BOOTINFO	0x0ff0

BOOTINFO *binfo = (BOOTINFO *)ADR_BOOTINFO;

struct MOUSE_DEC {
	byte buf[3], phase;
	int x, y, button;
};

void wait_KBC_sendready();
void init_keyboard();
void enable_mouse(struct MOUSE_DEC *);
int mouse_decode(struct MOUSE_DEC *, byte);

void Main() {
	char s[30], keyboardbuf[32], mousebuf[128];
	int i;
	int mx = (binfo->scrnx-16)/2, my = (binfo->scrny-28-16)/2;
	struct MOUSE_DEC mdec;

	init_gdtidt();	// GDT/IDTの初期化
	init_pic();		// PICの初期化
	io_sti();		// 割り込み許可

	// キーボード用QUEUEの初期化
	queue_init(&qKeyboard, 32, keyboardbuf);
	// マウス用QUEUEの初期化
	queue_init(&qMouse, 128, mousebuf);

	io_out8(PIC0_IMR, 0xf9);	// PIC1とキーボードを許可(0xf9)
	io_out8(PIC1_IMR, 0xef);	// マウスを許可

	init_keyboard();
	init_palete();	// パレットの初期化
	init_screen(binfo->scrnx, binfo->scrny);
	draw_mouse(mx, my);

	enable_mouse(&mdec);

	while(1) {
		io_cli();
		// キーボードとマウスの両方からデータを受け取っていない
		if(queue_status(&qKeyboard) + queue_status(&qMouse) == 0) {
			io_stihlt();
		// キーボードかマウスからデータを受け取っている
		} else {
			// キーボードからデータを受信
			if(queue_status(&qKeyboard)) {
				i = queue_get(&qKeyboard);
				io_sti();
				boxfill(COL_008484, 0, 16, 15, 31);
				lsprintf(s, "%X", i);
				vprint(s, 0, 16, COL_FFFFFF);
			// マウスかデータを受信
			} else if(queue_status(&qMouse)) {
				i = queue_get(&qMouse);
				io_sti();
				// マウスデータが3バイト揃う
				if(mouse_decode(&mdec, i)) {
					// データの表示
					lsprintf(s, "[lcr] x=%d y=%d", mdec.x, mdec.y);
					if((mdec.button & 0x01) != 0) s[1] = 'L';
					if((mdec.button & 0x02) != 0) s[3] = 'R';
					if((mdec.button & 0x03) != 0) s[2] = 'C';
					boxfill(COL_008484, 32, 16, 32+30*8-1, 31);
					vprint(s, 32, 16, COL_FFFFFF);

					// マウスの移動
					// マウスカーソルの消去
					boxfill(COL_008484, mx, my, mx + 15, my + 15);
					mx += mdec.x; my += mdec.y;
					if(mx < 0) mx = 0;
					if(my < 0) my = 0;
					if(mx > binfo->scrnx - 16) mx = binfo->scrnx - 16;
					if(my > binfo->scrny - 16) my = binfo->scrny - 16;
					putblock(16, 16, mx, my, mcursor, 16);
				}
			}
		}
	}
}

// キーボードコントローラがデータ送信可能になるのを待つ
void wait_KBC_sendready() {
	for(;;) {
		if((io_in8(PORT_KEYSTA) & KEYSTA_SEND_NOTREADY) == 0) break;
	}
}

// キーボードコントローラの初期化
void init_keyboard() {
	wait_KBC_sendready();
	io_out8(PORT_KEYCMD, KEYCMD_WRITE_MODE);
	wait_KBC_sendready();
	io_out8(PORT_KEYDAT, KBC_MODE);
}

// マウスの有効化
void enable_mouse(struct MOUSE_DEC *mdec) {
	wait_KBC_sendready();
	io_out8(PORT_KEYCMD, KEYCMD_SENDTO_MOUSE);
	wait_KBC_sendready();
	io_out8(PORT_KEYDAT, MOUSECMD_ENABLE);
	// 成功すればACK(0xfa)を返却
	mdec->phase = 0;
}

int mouse_decode(struct MOUSE_DEC *mdec, byte dat) {
	if(mdec->phase == 0) {
		if(dat == 0xfa) mdec->phase = 1;
		return 0;
	}
	
	if(mdec->phase == 1) {
		if((dat & 0xc8) == 0x08) {
			mdec->buf[0] = dat;
			mdec->phase = 2;
		}
		return 0;
	}
	
	if(mdec->phase == 2) {
		mdec->buf[1] = dat;
		mdec->phase = 3;
		return 0;
	}
	
	if(mdec->phase == 3) {
		mdec->buf[2] = dat;
		mdec->phase = 1;
		mdec->button = mdec->buf[0] & 0x07;
		mdec->x = mdec->buf[1];
		mdec->y = mdec->buf[2];
		if((mdec->buf[0] & 0x10) != 0) {
			mdec->x |= 0xffffff00;
		}
		if((mdec->buf[0] & 0x20) != 0) {
			mdec->y |= 0xffffff00;
		}
		mdec->y = - mdec->y;
		return 1;
	}

	return -1;
}
